using System.Collections;
using System.Collections.Generic;
using UnityEngine;

namespace VAMLaunchPlugin.MotionSources
{

    public class ZoneSimulatorOptions
    {
        public float CurrentTargetPos; // get/set
        public float PositionSampleRate; //get
        public float MinLaunchSignalTimeThreshold; //get
        public float MaxLaunchSignalTimeThreshold; //get
        public float CurrentLaunchSignalTimeThreshold; //get/set
        public float LowerVelocityBarrier; //get
        public float HigherVelocityBarrier; //get
        public float LaunchSpeedMultiplier; //get
        public float InputMultiplier; //get
        public float MinPosition; //get
        public float MaxPosition; //get
        
        public float MinDistanceBeforePrediction; //get
        public float MaxDistanceBeforePrediction; //get
        public float MaxDistancePercentBeforePrediction; //get
        public int MaxDistanceBeforePredictionMode; //get

        public int VelocitySmoothingSteps; //get

        public float CurrentAxisMinPosition; //out
        public float CurrentAxisMaxPosition; //out

        public bool ReturnToSenderEnabled; //get
        public float ReturnToSenderTime; //get

        public float DynamicRangeMin; //get/set
        public float DynamicRangeMax; //get/set
        public bool DynamicRangeEnabled; //get

        public float CurrentSign; //out
    }

    public class ZoneSimulator : ILaunchMotionSource
    {

        public const int VELOCITY_BUFFER_CAPACITY = 15;
        public const int AVG_VELOCITY_BUFFER_CAPACITY = 80;

        private float _lastPosFactor;
        private float _lastMinPos = 0f;
        private float _lastMaxPos = 99f;

        private float _lastSentPosFactor;
        private float _lastSentPosTime;
        private float _positionSampleTimer;
        private float _timeMovingUpwards;
        private float _timeMovingDownwards;
        private bool _moveUpwardsTimeHit;
        private bool _moveDownwardsTimeHit;

        private bool _upAllowedToBeCleared;
        private bool _downAllowedToBeCleared;
        private bool _upFailsafeConsumed;
        private bool _downFailsafeConsumed;
        private UpDownTotalTime _downTotalTimes;
        private UpDownTotalTime _upTotalTimes;

        private class UpDownTotalTime
        {
            public const int LAST_TOTAL_TIMES_CAPACITY = 5;

            public float[] TotalTimes;
            public int Index;
            public bool InRecord;
            public float Average;

            public UpDownTotalTime()
            {
                TotalTimes = new float[LAST_TOTAL_TIMES_CAPACITY];
                for (int i = 0; i < LAST_TOTAL_TIMES_CAPACITY; i++)
                {
                    TotalTimes[i] = 1f;
                }
                InRecord = false;
                Average = 1f;
            }

            public void StartRecord()
            {
                if (InRecord)
                    return;

                InRecord = true;
                Index++;
                if (Index >= TotalTimes.Length)
                {
                    Index = 0;
                }
                TotalTimes[Index] = 0f;
            }
            public void StopRecord()
            {
                InRecord = false;
                Average = CalculateAverage();

                // if average speed goes UP (going from faster storkes to slower), make all past average go up quicker to avoid getting "stuck" in the fast-speed stroke.
                if ((TotalTimes[Index] * 0.7f) > Average)
                {
                    for (int i = 0; i < TotalTimes.Length; i++)
                    {
                        if (i == Index)
                            continue;

                        TotalTimes[i] = TotalTimes[Index] * 0.8f;
                    }
                }
            }
            public void AddToCurrentRecord(float time)
            {
                TotalTimes[Index] += time;
            }
            public float CalculateAverage()
            {
                float total = 0f;
                for(int i = 0; i < TotalTimes.Length; i++)
                {
                    if (i == Index)
                        continue;
                    total += TotalTimes[i];
                }
                return (total / (float)(TotalTimes.Length - 1));
            }
        }

        private Queue<float> _upwardsVelocityBuffer = new Queue<float>(VELOCITY_BUFFER_CAPACITY);
        private Queue<float> _downwardsVelocityBuffer = new Queue<float>(VELOCITY_BUFFER_CAPACITY);
        private List<float> _velocityHistory = new List<float>(AVG_VELOCITY_BUFFER_CAPACITY);

        public ZoneSimulatorOptions Options;

        public ZoneSimulator(ZoneSimulatorOptions optionsRef)
        {
            Options = optionsRef;
            _upTotalTimes = new UpDownTotalTime();
            _downTotalTimes = new UpDownTotalTime();
        }

        /// <param name="lastSimulatedPos">lastSimulatedPos in range 0 to 1</param>
        public bool OnUpdate(float lastSimulatedPos, ref byte outPos, ref byte outSpeed, float capMax, float capMaxSpeed)
        {
            bool shouldMoveLaunch = ProcessTarget(lastSimulatedPos, ref outPos, ref outSpeed, capMax, capMaxSpeed);

            return shouldMoveLaunch;
        }

        public void OnSimulatorUpdate(float prevPos, float newPos, float deltaTime)
        {
            // Do nothing
        }


        private float _currentMaxDynamicRange;
        private float _lastMaxPushTime;
        private float _currentMaxTempTime = 0f;
        private float _currentMaxTempRange;

        private float _currentMinDynamicRange;
        private float _lastMinPushTime;
        private float _currentMinTempTime = 0f;
        private float _currentMinTempRange;

        public void UpdateDynamicRange(float currentTargetPos, out float targetMinPosition, out float targetMaxPosition)
        {
            
            targetMinPosition = Options.MinPosition;
            targetMaxPosition = Options.MaxPosition;

            // dynamic is disabled
            return; 
            
            const float MIN_RANGE = 35f;
            const float TARGET_PADDING = 10f;
            const float DYNAMIC_CHANGE_SPEED = 8f; // range per second
            const float TIME_BEFORE_SHRINK = 3f; // seconds
            const float TEMP_INCREASE_PERIOD = 0.4f; // seconds, must be less than TIME_BEFORE_SHRINK
            const float TEMP_INCREASE_AMOUNT_RANGE = 15; // additional increase amount to push the min/max when range needs to expand
            
            const float SNAP_AMOUNT_MIN = 10f;
            const float SNAP_AMOUNT_MAX = 8f;

            if ((Options.MaxPosition - Options.MinPosition) < MIN_RANGE)
            {
                // not enough space for a dynamic zone
                _currentMinDynamicRange = Options.MinPosition;
                _currentMaxDynamicRange = Options.MaxPosition;
                Options.DynamicRangeMin = _currentMinDynamicRange;
                Options.DynamicRangeMax = _currentMaxDynamicRange;
                return;
            }

            // Update both min and max at the same time. If they create a range that's less than 40, expand the range to be at least 40!
            if (_currentMaxDynamicRange < (currentTargetPos + TARGET_PADDING))
            {
                _currentMaxDynamicRange = Mathf.Min(Options.MaxPosition, (currentTargetPos + TARGET_PADDING));
                _lastMaxPushTime = Time.time;
                _currentMaxTempTime = Time.time;
                float tempTarget = Mathf.Min(Options.MaxPosition, _currentMaxDynamicRange + TEMP_INCREASE_AMOUNT_RANGE);
                if (_currentMaxTempRange < (tempTarget - TEMP_INCREASE_AMOUNT_RANGE / 2f) || _currentMaxTempRange > (tempTarget + TEMP_INCREASE_AMOUNT_RANGE / 2f))
                {
                    _currentMaxTempRange = tempTarget;
                }
            }
            else if (Time.time - _lastMaxPushTime > TIME_BEFORE_SHRINK)
            {
                _currentMaxDynamicRange = Mathf.Max(Options.MinPosition, _currentMaxDynamicRange - Time.deltaTime * DYNAMIC_CHANGE_SPEED);
            }
            else if (Mathf.Abs(_currentMaxDynamicRange - currentTargetPos) <= TARGET_PADDING * 2f)
            {
                // reset pushTime as the threshold is close enough, no need to trigger it!
                _lastMaxPushTime = Time.time;
            }


            if (_currentMinDynamicRange > (currentTargetPos - TARGET_PADDING))
            {
                _currentMinDynamicRange = Mathf.Max(Options.MinPosition, (currentTargetPos - TARGET_PADDING));
                _lastMinPushTime = Time.time;
                _currentMinTempTime = Time.time;
                float tempTarget = Mathf.Max(Options.MinPosition, _currentMinDynamicRange - TEMP_INCREASE_AMOUNT_RANGE);
                if (_currentMinTempRange < (tempTarget - TEMP_INCREASE_AMOUNT_RANGE / 2f) || _currentMinTempRange > (tempTarget + TEMP_INCREASE_AMOUNT_RANGE / 2f))
                {
                    _currentMinTempRange = tempTarget;
                }
            }
            else if (Time.time - _lastMinPushTime > TIME_BEFORE_SHRINK)
            {
                _currentMinDynamicRange = Mathf.Min(Options.MaxPosition, (_currentMinDynamicRange + Time.deltaTime * DYNAMIC_CHANGE_SPEED));
            }
            else if (Mathf.Abs(_currentMinDynamicRange - currentTargetPos) <= TARGET_PADDING * 2f)
            {
                _lastMinPushTime = Time.time;
            }

            // Check expand the zone to cover min-range
            float forcedExpandedAmount = 0f;
            if ((_currentMaxDynamicRange - _currentMinDynamicRange) < MIN_RANGE)
            {
                forcedExpandedAmount = MIN_RANGE - (_currentMaxDynamicRange - _currentMinDynamicRange);
                _currentMinDynamicRange -= forcedExpandedAmount / 2f;
                _currentMaxDynamicRange += forcedExpandedAmount / 2f;
            }

            // check out-of-bounds and push back zone
            if (_currentMinDynamicRange < Options.MinPosition)
            {
                float toPush = Options.MinPosition - _currentMinDynamicRange;
                _currentMinDynamicRange = Options.MinPosition;
                _currentMaxDynamicRange += toPush;
            }
            if (_currentMaxDynamicRange > Options.MaxPosition)
            {
                float toPush = _currentMaxDynamicRange - Options.MaxPosition;
                _currentMaxDynamicRange = Options.MaxPosition;
                _currentMinDynamicRange -= toPush;
            }

            float finalDynamicMinRange = _currentMinDynamicRange;
            float finalDynamicMaxRange = _currentMaxDynamicRange;

            if ((Time.time - _currentMinTempTime) < TEMP_INCREASE_PERIOD && finalDynamicMinRange > _currentMinTempRange)
            {
                finalDynamicMinRange = _currentMinTempRange;
            }
            if ((Time.time - _currentMaxTempTime) < TEMP_INCREASE_PERIOD && finalDynamicMaxRange < _currentMaxTempRange)
            {
                finalDynamicMaxRange = _currentMaxTempRange;
            }
            

            // perform last-minute snapping to min/max to get the most out of the linear device
            bool minSnapped = false;
            if (finalDynamicMinRange < (Options.MinPosition + SNAP_AMOUNT_MIN))
            {
                float toSnap = finalDynamicMinRange - Options.MinPosition;
                finalDynamicMinRange = Options.MinPosition;
                minSnapped = true;
                // only allowed to snap end position maximum the expaned zone
                toSnap = Mathf.Max(forcedExpandedAmount / 2f, toSnap);
                finalDynamicMaxRange -= toSnap;
            }
            if (finalDynamicMaxRange > (Options.MaxPosition - SNAP_AMOUNT_MAX))
            {
                float toSnap = Options.MaxPosition - finalDynamicMaxRange;
                finalDynamicMaxRange = Options.MaxPosition;
                if (!minSnapped)
                {
                    toSnap = Mathf.Max(forcedExpandedAmount / 2f, toSnap);
                    finalDynamicMinRange += toSnap;
                }
            }

            if (Options.DynamicRangeEnabled)
            {
                targetMinPosition = finalDynamicMinRange;
                targetMaxPosition = finalDynamicMaxRange;
            }

            Options.DynamicRangeMin = finalDynamicMinRange;
            Options.DynamicRangeMax = finalDynamicMaxRange;
        }

        private class AxisMinMaxValue
        {
            public float Pending;
            public float Value;
            public float Push;
        }
        private AxisMinMaxValue AxisMin = new AxisMinMaxValue();
        private AxisMinMaxValue AxisMax = new AxisMinMaxValue();
        private AxisMinMaxValue MovingFromAxis = null;
        private AxisMinMaxValue LastTriggeredFromAxis = null;

        private void UpdateAxisMinMax()
        {
            const float PUSH = 10f;

            if (MovingFromAxis == null)
                MovingFromAxis = AxisMin;

            // Min
            if (AxisMin.Push >= Options.CurrentTargetPos + PUSH)
            {
                AxisMin.Pending = Options.CurrentTargetPos;
                AxisMin.Push = Options.CurrentTargetPos + PUSH;
                if (AxisMin.Value > AxisMin.Pending)
                    AxisMin.Value = AxisMin.Pending;

                MovingFromAxis = AxisMax;
            }
            else if (AxisMin.Push < Options.CurrentTargetPos)
            {
                AxisMin.Value = AxisMin.Pending;
                AxisMin.Push = Options.CurrentTargetPos;
            }

            // Max
            if (AxisMax.Push <= Options.CurrentTargetPos - PUSH)
            {
                AxisMax.Pending = Options.CurrentTargetPos;
                AxisMax.Push = Options.CurrentTargetPos - PUSH;
                if (AxisMax.Value < AxisMax.Pending)
                    AxisMax.Value = AxisMax.Pending;

                MovingFromAxis = AxisMin;
            }
            else if (AxisMax.Push > Options.CurrentTargetPos)
            {
                AxisMax.Value = AxisMax.Pending;
                AxisMax.Push = Options.CurrentTargetPos;
            }

            Options.CurrentAxisMinPosition = AxisMin.Value;
            Options.CurrentAxisMaxPosition = AxisMax.Value;
        }

        /// <param name="lastSimulatedPos">lastSimulatedPos in range 0 to 1</param>
        private bool ProcessTarget(float lastSimulatedPos, ref byte outPos, ref byte outSpeed, float capMax, float capMaxSpeed)
        {
            bool shouldMoveLaunch = false;

            _positionSampleTimer -= Time.deltaTime;
            if (_positionSampleTimer > 0.0f)
            {
                return false;
            }

            var sampleInterval = GetPositionSampleInterval();
            _positionSampleTimer = sampleInterval - Mathf.Min(sampleInterval, -_positionSampleTimer);

            // Convert target position into zone local space (-1.0 to 1.0 (from 0.0 to 1.0)) - for easy convert of existing code
            lastSimulatedPos = ((lastSimulatedPos * 2f) - 1f) * Options.InputMultiplier;
            Vector3 localZonePos = new Vector3(lastSimulatedPos, lastSimulatedPos, lastSimulatedPos);
            if (localZonePos.y < -1.0f)
                localZonePos.y = -1.0f;
            if (localZonePos.y > 1.0f)
                localZonePos.y = 1.0f;

            // Convert Y range from -1.0, 1.0 to 0.0f, 99.0f
            float posFactor = Mathf.InverseLerp(-1.0f, 1.0f, localZonePos.y);
            Options.CurrentTargetPos = (Mathf.Lerp(0f, LaunchUtils.LAUNCH_MAX_VAL, posFactor));

            float targetPosDeltaSinceLastSent = Options.CurrentTargetPos - _lastSentPosFactor;

            //float dynamicMinPosition;
            //float dynamicMaxPosition;
            //UpdateDynamicRange(Options.CurrentTargetPos, out dynamicMinPosition, out dynamicMaxPosition);

            //float minPos = System.Math.Min(System.Math.Max(0, dynamicMinPosition), LaunchUtils.LAUNCH_MAX_VAL);
            //float maxPos = System.Math.Max(0f, System.Math.Min(dynamicMaxPosition, LaunchUtils.LAUNCH_MAX_VAL));

            float minPos = System.Math.Min(System.Math.Max(0, Options.MinPosition), LaunchUtils.LAUNCH_MAX_VAL);
            float maxPos = System.Math.Max(0f, System.Math.Min(Options.MaxPosition, LaunchUtils.LAUNCH_MAX_VAL));

            float maxSpeed = LaunchUtils.LAUNCH_MAX_SPEED;

            // cap limits
            if (capMax > 0f && maxPos > capMax)
            {
                maxPos = capMax;
            }
            if (capMaxSpeed > LaunchUtils.LAUNCH_MIN_SPEED && maxSpeed > capMaxSpeed)
            {
                maxSpeed = capMaxSpeed;
            }

            // scale posFactor to current set min/max position to get "scale" the input 0-1 to the targeted device range min-max and to get calculated speeds to scale.
            // however - doing this also means that when min/max changes - so does the current posFactor which in turn triggers calculations to occure with janky output.
            posFactor = (Mathf.Lerp(minPos, maxPos, posFactor));

            if (_lastMinPos != minPos || _lastMaxPos != maxPos)
            {
                if (_lastMaxPos > (_lastMinPos + 1f))
                {
                    float p = (_lastPosFactor - _lastMinPos) / (_lastMaxPos - _lastMinPos);
                    _lastMinPos = minPos;
                    _lastMaxPos = maxPos;
                    float adjustedLastPosFactor = minPos + (maxPos - minPos) * p;
                    _lastPosFactor = adjustedLastPosFactor;
                } 
                else
                {
                    _lastMinPos = minPos;
                    _lastMaxPos = maxPos;
                }
            }

            // float launchFinalPercent = Mathf.Lerp(0, LaunchUtils.LAUNCH_MAX_VAL, posFactor);

            // Calculate required launch move speed in order to reach new position
            float speed = LaunchUtils.PredictMoveSpeed(_lastPosFactor, posFactor, Time.deltaTime);
            float launchDynamicSpeedModifier = 1f;
            if (Options.DynamicRangeEnabled)
            {
                launchDynamicSpeedModifier = 1f / ( (maxPos - minPos) / (Options.MaxPosition - Options.MinPosition) );
            }
            
            // We store this speed into our velocity buffer to help with generating a rolling average speed.
            if (_velocityHistory.Count == AVG_VELOCITY_BUFFER_CAPACITY)
            {
                _velocityHistory.RemoveAt(0);
            }
            _velocityHistory.Add(speed);

            float delta = posFactor - _lastPosFactor;
            if (Mathf.Abs(delta) < 0.02f)
            {
                return false;
            }

            bool backToSenderPossible = Options.ReturnToSenderEnabled && (Time.time - _lastSentPosTime) > Mathf.Max(Options.ReturnToSenderTime, Options.CurrentLaunchSignalTimeThreshold + 0.15f);

            int deltaSign = delta > 0.0f ? 1 : -1;
            Options.CurrentSign = deltaSign;

            UpdateAxisMinMax();

            // We apply the users desired speed multiplier (After we store the original in for the rolling average to
            // prevent throwing off all the other parameters that tune the prediction logic)
            speed = Mathf.Clamp(speed * Options.LaunchSpeedMultiplier, 0.0f, LaunchUtils.LAUNCH_MAX_VAL);

            bool movedForcedAmount = false;
            bool movedMinimumAmount = Mathf.Abs(targetPosDeltaSinceLastSent) >= Options.MinDistanceBeforePrediction;
            
            if (movedMinimumAmount)
            {
                if (LastTriggeredFromAxis != MovingFromAxis)
                {
                    if (Options.MaxDistanceBeforePredictionMode == 2)
                    {
                        float whole = Mathf.Max((AxisMax.Value - AxisMin.Value), Options.MinDistanceBeforePrediction + 5f); // min range allowed: MinDistanceBeforePrediction + 5
                        movedForcedAmount = (Mathf.Abs(MovingFromAxis.Value - Options.CurrentTargetPos) / whole) > Options.MaxDistancePercentBeforePrediction * 0.01f;
                    }
                    else if (Options.MaxDistanceBeforePredictionMode == 1)
                    {
                        movedForcedAmount = Mathf.Abs(MovingFromAxis.Value - Options.CurrentTargetPos) > Options.MaxDistanceBeforePrediction;
                    }

                    if (movedForcedAmount)
                    {
                        LastTriggeredFromAxis = MovingFromAxis;
                    }
                }
            }

            // Are we moving up?
            if (deltaSign == 1)
            {
                // Storing a rolling velocity for upwards motion
                if (_upwardsVelocityBuffer.Count == VELOCITY_BUFFER_CAPACITY)
                {
                    _upwardsVelocityBuffer.Dequeue();
                }
                _upwardsVelocityBuffer.Enqueue(speed);
                _upTotalTimes.StartRecord();
                _downTotalTimes.StopRecord();

                _upTotalTimes.AddToCurrentRecord(Time.deltaTime);

                // If we continue moving in the same direction then we accumulate time, if that time goes over our
                // calculated threshold then we should consider telling the Launch about this movement
                var prevTime = _timeMovingUpwards;
                _timeMovingUpwards += Time.deltaTime;

                _moveUpwardsTimeHit |= prevTime <= Options.CurrentLaunchSignalTimeThreshold && _timeMovingUpwards >= Options.CurrentLaunchSignalTimeThreshold;

                // Failsafe when storkes in animation changes to extreme slow or fast speeds, force the launch to move instead freezing (which is does for most settings)
                bool shouldFailsafe = !_upFailsafeConsumed && _upTotalTimes.Average < Options.CurrentLaunchSignalTimeThreshold && _timeMovingUpwards > _upTotalTimes.Average * 0.5f && movedMinimumAmount;

                if (_moveUpwardsTimeHit && movedMinimumAmount ||
                    _timeMovingUpwards >= 0.4f && movedMinimumAmount ||
                    movedForcedAmount ||
                    shouldFailsafe) 
                {
                    _moveUpwardsTimeHit = false;
                    _upFailsafeConsumed = shouldFailsafe;

                    // if (movedForcedAmount)
                    // SuperController.LogMessage("Up " + movedForcedAmount.ToString());

                    // We take the highest speed during this upwards motion as it is most likely the highest speed that
                    // will be the closest representation of the motion.
                    float highestSpeed = RetrieveHighestSpeed(_upwardsVelocityBuffer);
                    highestSpeed = Mathf.Round(highestSpeed * launchDynamicSpeedModifier);

                    // Only if the speed is small do we tell the launch to go to exact locations 
                    // (Not happy with this, need to think of a better way to deal with slower motions)
                    float launchPos = highestSpeed > 1.0f ? maxPos : posFactor;
                    float launchSpeed = Mathf.Min(highestSpeed, maxSpeed);

                    // Tell the Launch to do it's thing!
                    shouldMoveLaunch = true;
                    outPos = (byte)launchPos;
                    outSpeed = (byte)launchSpeed;

                    UpdateSignalThreshold();

                    if (highestSpeed < 1.0f)
                    {
                        _timeMovingUpwards = 0.0f;
                        _upwardsVelocityBuffer.Clear();
                    }

                    _lastSentPosFactor = Options.CurrentTargetPos;
                    _lastSentPosTime = Time.time;

                    // flag that up values can be cleared
                    _upAllowedToBeCleared = true;
                }

                if (_downAllowedToBeCleared)
                {
                    _timeMovingDownwards = 0.0f;
                    _downwardsVelocityBuffer.Clear();
                    _downAllowedToBeCleared = false;
                }
                _downFailsafeConsumed = false;
            }
            // This does the same as above except for downwards motion
            else if (deltaSign == -1)
            {
                if (_downwardsVelocityBuffer.Count == VELOCITY_BUFFER_CAPACITY)
                {
                    _downwardsVelocityBuffer.Dequeue();
                }
                _downwardsVelocityBuffer.Enqueue(speed);
                _downTotalTimes.StartRecord();
                _upTotalTimes.StopRecord();

                _downTotalTimes.AddToCurrentRecord(Time.deltaTime);

                var prevTime = _timeMovingDownwards;
                _timeMovingDownwards += Time.deltaTime;

                _moveDownwardsTimeHit |= prevTime <= Options.CurrentLaunchSignalTimeThreshold && _timeMovingDownwards >= Options.CurrentLaunchSignalTimeThreshold;

                bool shouldFailsafe = !_downFailsafeConsumed && _downTotalTimes.Average < Options.CurrentLaunchSignalTimeThreshold && _timeMovingDownwards > _downTotalTimes.Average * 0.5f && movedMinimumAmount;

                if (_moveDownwardsTimeHit && movedMinimumAmount ||
                    _timeMovingDownwards >= 0.4f && movedMinimumAmount ||
                    movedForcedAmount ||
                    shouldFailsafe)
                {
                    _moveDownwardsTimeHit = false;
                    _downFailsafeConsumed = shouldFailsafe;

                    // if (movedForcedAmount)
                    // SuperController.LogMessage("Down " + movedForcedAmount.ToString());

                    float highestSpeed = RetrieveHighestSpeed(_downwardsVelocityBuffer);
                    highestSpeed = Mathf.Round(highestSpeed * launchDynamicSpeedModifier);

                    float launchPos = highestSpeed > 1.0f ? minPos : posFactor;
                    float launchSpeed = Mathf.Min(highestSpeed, maxSpeed);

                    // Tell the Launch to do it's thing!
                    shouldMoveLaunch = true;
                    outPos = (byte)launchPos;
                    outSpeed = (byte)launchSpeed;

                    UpdateSignalThreshold();

                    if (highestSpeed < 1.0f)
                    {
                        _timeMovingDownwards = 0.0f;
                        _downwardsVelocityBuffer.Clear();
                    }

                    _lastSentPosFactor = Options.CurrentTargetPos;
                    _lastSentPosTime = Time.time;

                    // flag that down values can be cleared
                    _downAllowedToBeCleared = true;
                }

                if (_upAllowedToBeCleared)
                {
                    _timeMovingUpwards = 0.0f;
                    _upwardsVelocityBuffer.Clear();
                    _upAllowedToBeCleared = false;
                }
                _upFailsafeConsumed = false;
            }

            _lastPosFactor = posFactor;

            if (!shouldMoveLaunch && backToSenderPossible && !movedMinimumAmount)
            {
                outPos = (byte)posFactor;
                outSpeed = (byte)LaunchUtils.LAUNCH_MIN_SPEED_CALC;
                shouldMoveLaunch = true;
                _lastSentPosTime = Time.time;
                _lastSentPosFactor = Options.CurrentTargetPos;
                // SuperController.LogMessage("Back To Sender");
            }

            return shouldMoveLaunch;
        }

        private float CalculateAverageSpeed()
        {
            var averageVel = 0.0f;
            for (int i = 0; i < _velocityHistory.Count; i++)
            {
                averageVel += _velocityHistory[i];
            }

            if (_velocityHistory.Count > 0)
            {
                averageVel /= _velocityHistory.Count;
            }

            return averageVel;
        }

        // The signal threshold is how much time must moving in a single direction before the system will consider it
        // a valid motion to tell the launch about.
        private void UpdateSignalThreshold()
        {
            var averageVel = CalculateAverageSpeed();

            // To calculate this threshold we take the average speed and use it to blend between two min,max threshold
            // values, if the average speed is higher we want the threshold to be shorter, and if slower we want the threshold
            // to be longer to give the best chance of choosing a good velocity.
            Options.CurrentLaunchSignalTimeThreshold = Mathf.Lerp(
                Options.MaxLaunchSignalTimeThreshold,
                Options.MinLaunchSignalTimeThreshold,
                Mathf.InverseLerp(
                    Options.LowerVelocityBarrier, 
                    Options.HigherVelocityBarrier, 
                    averageVel)
                );
        }

        private float RetrieveHighestSpeed(Queue<float> speeds)
        {
            float highest1 = 0f;
            float highest2 = 0f;
            float highest3 = 0f;
            while (speeds.Count > 0)
            {
                var spd = speeds.Dequeue();
                if (spd > highest1)
                {
                    if (highest1 < 1f)
                    {
                        highest1 = spd;
                        highest2 = spd;
                        highest3 = spd;
                    }
                    else
                    {
                        highest3 = highest2;
                        highest2 = highest1;
                        highest1 = spd;
                    }
                } 
                else if (spd > highest2)
                {
                    highest3 = highest2;
                    highest2 = spd;
                }
                else if (spd > highest3)
                {
                    highest3 = spd;
                }
            }

            switch (Options.VelocitySmoothingSteps)
            {
                case 1: return highest1;
                case 2: return (highest1 + highest2) / 2f;
                case 3: return (highest1 + highest2 + highest3) / 3f;
            }

            return (highest1 + highest2 + highest3) / 3f;
        }

        private float GetPositionSampleInterval()
        {
            return 1.0f / Options.PositionSampleRate;
        }
    }

}