using DebugUtils;
using System;
using System.Linq;
using System.Text;
using ToySerialController.MotionSource;
using ToySerialController.Device.OutputTarget;
using UnityEngine;

namespace ToySerialController
{
    public partial class TCodeDevice : IDevice
    {
        protected readonly float[] XTarget, RTarget, ETarget;
        protected readonly float[] XCmd, RCmd, ECmd;
        protected readonly float[] LastXCmd, LastRCmd, LastECmd;
        private readonly StringBuilder _stringBuilder;

        private float? _lastNoCollisionTime;
        private bool _lastNoCollisionSmoothingEnabled;
        private float _lastCollisionSmoothingT;
        private float _lastNoCollisionSmoothingStartTime, _lastNoCollisionSmoothingDuration;
        private bool _isLoading;

        protected string DeviceReport { get; set; }
        public string GetDeviceReport() => DeviceReport;

        public static float TargetSwappedTime;
        public static float C0_RANGE_LIMIT_VALUE = 0f;

        public TCodeDevice()
        {
            XTarget = new float[3];
            RTarget = new float[3];
            ETarget = new float[9];

            XCmd = new float[] { 0.5f, 0.5f, 0.5f };
            RCmd = new float[] { 0.5f, 0.5f, 0.5f };
            ECmd = new float[9];

            LastXCmd = new float[] { float.NaN, float.NaN, float.NaN };
            LastRCmd = new float[] { float.NaN, float.NaN, float.NaN };
            LastECmd = Enumerable.Range(0, 9).Select(_ => float.NaN).ToArray();

            _lastNoCollisionTime = Time.time;
            _stringBuilder = new StringBuilder();
            TargetSwappedTime = 0f;
        }

        private string AppendIfChanged(StringBuilder stringBuilder, string axisName, float cmd, ref float lastCmd)
        {
            if (!float.IsNaN(lastCmd) && Mathf.Abs(lastCmd - cmd) * 999 < 1)
            {
                //if ((cmd != 0f && cmd != 1f) || (lastCmd == cmd)) // enable to force 0f and 1f values - but not required!
                //{
                return string.Empty;
                //}
            }

            lastCmd = cmd;
            var command = $"{axisName}{Mathf.RoundToInt(Mathf.Clamp01(cmd) * 9999):0000}I{Mathf.RoundToInt(Time.deltaTime * 1000)}";
            stringBuilder.Append(command).Append(" ");
            return command;
        }

        public bool Update(IMotionSource motionSource, IOutputTarget outputTarget)
        {
            bool tracking = false;
            if (_isLoading)
            {
                for (var i = 0; i < 9; i++)
                    ETarget[i] = Mathf.Lerp(ETarget[i], 0f, 0.05f);

                for (var i = 0; i < 3; i++)
                {
                    XTarget[i] = Mathf.Lerp(XTarget[i], 0.5f, 0.05f);
                    RTarget[i] = Mathf.Lerp(RTarget[i], 0f, 0.05f);
                }
            }
            else if(motionSource != null)
            {
                tracking = UpdateMotion(motionSource) >= 0f;

                DebugDraw.DrawCircle(motionSource.TargetPosition + motionSource.TargetUp * RangeMinL0Slider.val * motionSource.ReferenceLength, motionSource.TargetUp, motionSource.TargetRight, Color.white, 0.05f);
                DebugDraw.DrawCircle(motionSource.TargetPosition + motionSource.TargetUp * RangeMaxL0Slider.val * motionSource.ReferenceLength, motionSource.TargetUp, motionSource.TargetRight, Color.white, 0.05f);
            }

            UpdateValues(outputTarget, tracking);

            return true;
        }        

        public void UpdateValues(IOutputTarget outputTarget, bool tracking)
        {
            if (_lastNoCollisionSmoothingEnabled)
            {
                _lastCollisionSmoothingT = Mathf.Pow(2, 10 * ((Time.time - _lastNoCollisionSmoothingStartTime) / _lastNoCollisionSmoothingDuration - 1));
                if (_lastCollisionSmoothingT > 1.0f)
                {
                    _lastNoCollisionSmoothingEnabled = false;
                    _lastCollisionSmoothingT = 0;
                }
            }

            UpdateL0(); UpdateL1(); UpdateL2();
            UpdateR0(); UpdateR1(); UpdateR2();
            UpdateV0();
            UpdateA0(); UpdateA1(); UpdateA2();
            UpdateC0();
            CheckNaN();

            _stringBuilder.Length = 0;
            var l0 = AppendIfChanged(_stringBuilder, "L0", XCmd[0], ref LastXCmd[0]);
            var l1 = AppendIfChanged(_stringBuilder, "L1", XCmd[1], ref LastXCmd[1]);
            var l2 = AppendIfChanged(_stringBuilder, "L2", XCmd[2], ref LastXCmd[2]);
            var r0 = AppendIfChanged(_stringBuilder, "R0", RCmd[0], ref LastRCmd[0]);
            var r1 = AppendIfChanged(_stringBuilder, "R1", RCmd[1], ref LastRCmd[1]);
            var r2 = AppendIfChanged(_stringBuilder, "R2", RCmd[2], ref LastRCmd[2]);
            var v0 = AppendIfChanged(_stringBuilder, "V0", ECmd[0], ref LastECmd[0]);
            var a0 = AppendIfChanged(_stringBuilder, "A0", ECmd[1], ref LastECmd[1]);
            var a1 = AppendIfChanged(_stringBuilder, "A1", ECmd[2], ref LastECmd[2]);
            var a2 = AppendIfChanged(_stringBuilder, "A2", ECmd[3], ref LastECmd[3]);
            var c0 = AppendIfChanged(_stringBuilder, "C0", ECmd[4], ref LastECmd[4]);

            var data = $"{_stringBuilder}\n";
            if (!string.IsNullOrEmpty(data.Trim()))
            {
                outputTarget?.Write(data);
            }

            outputTarget?.Update(tracking && !_isLoading);

            _stringBuilder.Length = 0;
            _stringBuilder.Append("    Target    Cmd    Output\n");
            _stringBuilder.Append("L0\t").AppendFormat("{0,5:0.00}", XTarget[0]).Append(",\t").AppendFormat("{0,5:0.00}", XCmd[0]).Append(",\t").AppendLine(l0);
            _stringBuilder.Append("L1\t").AppendFormat("{0,5:0.00}", XTarget[1]).Append(",\t").AppendFormat("{0,5:0.00}", XCmd[1]).Append(",\t").AppendLine(l1);
            _stringBuilder.Append("L2\t").AppendFormat("{0,5:0.00}", XTarget[2]).Append(",\t").AppendFormat("{0,5:0.00}", XCmd[2]).Append(",\t").AppendLine(l2);
            _stringBuilder.Append("R0\t").AppendFormat("{0,5:0.00}", RTarget[0]).Append(",\t").AppendFormat("{0,5:0.00}", RCmd[0]).Append(",\t").AppendLine(r0);
            _stringBuilder.Append("R1\t").AppendFormat("{0,5:0.00}", RTarget[1]).Append(",\t").AppendFormat("{0,5:0.00}", RCmd[1]).Append(",\t").AppendLine(r1);
            _stringBuilder.Append("R2\t").AppendFormat("{0,5:0.00}", RTarget[2]).Append(",\t").AppendFormat("{0,5:0.00}", RCmd[2]).Append(",\t").AppendLine(r2);
            _stringBuilder.Append("V0\t").AppendFormat("{0,5:0.00}", ETarget[0]).Append(",\t").AppendFormat("{0,5:0.00}", ECmd[0]).Append(",\t").AppendLine(v0);
            _stringBuilder.Append("A0\t").AppendFormat("{0,5:0.00}", ETarget[1]).Append(",\t").AppendFormat("{0,5:0.00}", ECmd[1]).Append(",\t").AppendLine(a0);
            _stringBuilder.Append("A1\t").AppendFormat("{0,5:0.00}", ETarget[2]).Append(",\t").AppendFormat("{0,5:0.00}", ECmd[2]).Append(",\t").AppendLine(a1);
            _stringBuilder.Append("A2\t").AppendFormat("{0,5:0.00}", ETarget[3]).Append(",\t").AppendFormat("{0,5:0.00}", ECmd[3]).Append(",\t").AppendLine(a2);
            _stringBuilder.Append("C0\t").AppendFormat("{0,5:0.00}", ETarget[4]).Append(",\t").AppendFormat("{0,5:0.00}", ECmd[4]).Append(",\t").Append(c0);
            DeviceReport = _stringBuilder.ToString();
        }

        public void CheckNaN()
        {
            SetDefaultWhereInvalidFloat(XCmd);
            SetDefaultWhereInvalidFloat(RCmd);
            SetDefaultWhereInvalidFloat(ECmd);
            SetDefaultWhereInvalidFloat(XTarget);
            SetDefaultWhereInvalidFloat(RTarget);
            SetDefaultWhereInvalidFloat(ETarget);
            SetDefaultWhereInvalidFloat(LastECmd);
            SetDefaultWhereInvalidFloat(LastRCmd);
            SetDefaultWhereInvalidFloat(LastXCmd);
        }

        public void SetDefaultWhereInvalidFloat(float[] arr)
        {
            for (int i = 0; i < arr.Length; i++)
            {
                if (float.IsNaN(arr[i]) || float.IsInfinity(arr[i]))
                {
                    arr[i] = 0f;
                }
            }
        }

        public void UpdateL0()
        {
            var t = Mathf.Clamp01((XTarget[0] - RangeMinL0Slider.val) / (RangeMaxL0Slider.val - RangeMinL0Slider.val));
            var output = Mathf.Clamp01(Mathf.Lerp(OutputMinL0Slider.val, OutputMaxL0Slider.val, t));

            if (InvertL0Toggle.val) output = 1f - output;
            if (EnableOverrideL0Toggle.val) output = OverrideL0Slider.val;
            if (_lastNoCollisionSmoothingEnabled)
                output = Mathf.Lerp(XCmd[0], output, _lastCollisionSmoothingT);

            XCmd[0] = Mathf.Lerp(XCmd[0], output, 1 - SmoothingSlider.val);
        }

        public void UpdateL1()
        {
            var t = Mathf.Clamp01((XTarget[1] + RangeMaxL1Slider.val) / (2 * RangeMaxL1Slider.val));
            var output = Mathf.Clamp01(OffsetL1Slider.val + 0.5f + Mathf.Lerp(-OutputMaxL1Slider.val, OutputMaxL1Slider.val, t));

            if (InvertL1Toggle.val) output = 1f - output;
            if (EnableOverrideL1Toggle.val) output = OverrideL1Slider.val;
            if (_lastNoCollisionSmoothingEnabled)
                output = Mathf.Lerp(XCmd[1], output, _lastCollisionSmoothingT);

            XCmd[1] = Mathf.Lerp(XCmd[1], output, 1 - SmoothingSlider.val);
        }

        public void UpdateL2()
        {
            var t = Mathf.Clamp01((XTarget[2] + RangeMaxL2Slider.val) / (2 * RangeMaxL2Slider.val));
            var output = Mathf.Clamp01(OffsetL2Slider.val + 0.5f + Mathf.Lerp(-OutputMaxL2Slider.val, OutputMaxL2Slider.val, t));

            if (InvertL2Toggle.val) output = 1f - output;
            if (EnableOverrideL2Toggle.val) output = OverrideL2Slider.val;
            if (_lastNoCollisionSmoothingEnabled)
                output = Mathf.Lerp(XCmd[2], output, _lastCollisionSmoothingT);

            XCmd[2] = Mathf.Lerp(XCmd[2], output, 1 - SmoothingSlider.val);
        }

        public void UpdateR0()
        {
            var t = Mathf.Clamp01(0.5f + (RTarget[0] / 2) / (RangeMaxR0Slider.val / 180));
            var output = Mathf.Clamp01(OffsetR0Slider.val + 0.5f + Mathf.Lerp(-OutputMaxR0Slider.val, OutputMaxR0Slider.val, t));

            if (InvertR0Toggle.val) output = 1f - output;
            if (EnableOverrideR0Toggle.val) output = OverrideR0Slider.val;
            if (_lastNoCollisionSmoothingEnabled)
                output = Mathf.Lerp(RCmd[0], output, _lastCollisionSmoothingT);

            RCmd[0] = Mathf.Lerp(RCmd[0], output, 1 - SmoothingSlider.val);
        }

        public void UpdateR1()
        {
            var t = Mathf.Clamp01(0.5f + (RTarget[1] / 2) / (RangeMaxR1Slider.val / 90));
            var output = Mathf.Clamp01(OffsetR1Slider.val + 0.5f + Mathf.Lerp(-OutputMaxR1Slider.val, OutputMaxR1Slider.val, t));

            if (InvertR1Toggle.val) output = 1f - output;
            if (EnableOverrideR1Toggle.val) output = OverrideR1Slider.val;
            if (_lastNoCollisionSmoothingEnabled)
                output = Mathf.Lerp(RCmd[1], output, _lastCollisionSmoothingT);

            RCmd[1] = Mathf.Lerp(RCmd[1], output, 1 - SmoothingSlider.val);
        }

        public void UpdateR2()
        {
            var t = Mathf.Clamp01(0.5f + (RTarget[2] / 2) / (RangeMaxR2Slider.val / 90));
            var output = Mathf.Clamp01(OffsetR2Slider.val + 0.5f + Mathf.Lerp(-OutputMaxR2Slider.val, OutputMaxR2Slider.val, t));

            if (InvertR2Toggle.val) output = 1f - output;
            if (EnableOverrideR2Toggle.val) output = OverrideR2Slider.val;
            if (_lastNoCollisionSmoothingEnabled)
                output = Mathf.Lerp(RCmd[2], output, _lastCollisionSmoothingT);

            RCmd[2] = Mathf.Lerp(RCmd[2], output, 1 - SmoothingSlider.val);
        }

        public void UpdateV0()
        {
            var output = Mathf.Clamp01(ETarget[0]);

            if (EnableOverrideV0Toggle.val) output = OverrideV0Slider.val;
            if (_lastNoCollisionSmoothingEnabled)
                output = Mathf.Lerp(ECmd[0], output, _lastCollisionSmoothingT);

            ECmd[0] = Mathf.Lerp(ECmd[0], output, 1 - SmoothingSlider.val);
        }

        public void UpdateA0()
        {
            var output = Mathf.Clamp01(ETarget[1]);

            if (EnableOverrideA0Toggle.val) output = OverrideA0Slider.val;
            if (_lastNoCollisionSmoothingEnabled)
                output = Mathf.Lerp(ECmd[1], output, _lastCollisionSmoothingT);

            ECmd[1] = Mathf.Lerp(ECmd[1], output, 1 - SmoothingSlider.val);
        }

        public void UpdateA1()
        {
            var output = Mathf.Clamp01(ETarget[2]);

            if (EnableOverrideA1Toggle.val) output = OverrideA1Slider.val;
            if (_lastNoCollisionSmoothingEnabled)
                output = Mathf.Lerp(ECmd[2], output, _lastCollisionSmoothingT);

            ECmd[2] = Mathf.Lerp(ECmd[2], output, 1 - SmoothingSlider.val);
        }

        public void UpdateA2()
        {
            var output = Mathf.Clamp01(ETarget[3]);

            if (EnableOverrideA2Toggle.val) output = OverrideA2Slider.val;
            if (_lastNoCollisionSmoothingEnabled)
                output = Mathf.Lerp(ECmd[3], output, _lastCollisionSmoothingT);

            ECmd[3] = Mathf.Lerp(ECmd[3], output, 1 - SmoothingSlider.val);
        }

        public void UpdateC0()
        {
            // L0 = XCmd[0]
            // R1 = RCmd[1]
            // R2 = RCmd[2]

            _l0VT.Update(XCmd[0]);
            _r1VT.Update(RCmd[1]);
            _r2VT.Update(RCmd[2]);
            _r12S.Update(_r1VT.Magnitude, _r2VT.Magnitude);

            // float target = XCmd[0] + SX(XCmd[0]) * RCmd[2] * (1f - XCmd[0]) * 0.75f;
            float p = SX(XCmd[0]) * SM(_l0VT.Magnitude, 0.4f);
            C0_RANGE_LIMIT_VALUE = p;
            float r1 = _r1VT.Value * _r12S.PerA;
            float r2 = _r2VT.Value * _r12S.PerB;

            // SuperController.LogMessage(" - p: " + p.ToString("0.000"));
            // SuperController.LogMessage(" - L0 M: " + _l0VT.Magnitude.ToString("0.000") + " - L0 V: " + _l0VT.Value.ToString("0.000") + " - L0 L/H" + _l0VT.Low.ToString("0.000") + " / " + _l0VT.High.ToString("0.000"));
            // SuperController.LogMessage(" - p: " + p.ToString("0.000") + " - PerA/B: " + _r12S.PerA.ToString("0.000") + " / " + _r12S.PerB.ToString("0.000") + " - R1/2: " + _r1VT.Value.ToString("0.000") + " / " + _r2VT.Value.ToString("0.000") + " - RM1/2: " + _r1VT.Magnitude.ToString("0.000") + " / " + _r2VT.Magnitude.ToString("0.000"));

            float target = XCmd[0] + r1 * p + r2 * p;
            var t = Mathf.Clamp01((target - RangeMinL0Slider.val) / (RangeMaxL0Slider.val - RangeMinL0Slider.val));
            var output = Mathf.Clamp01(Mathf.Lerp(OutputMinL0Slider.val, OutputMaxL0Slider.val, t));

            if (EnableOverrideL0Toggle.val) output = OverrideL0Slider.val;
            if (_lastNoCollisionSmoothingEnabled)
                output = Mathf.Lerp(ECmd[4], output, _lastCollisionSmoothingT);

            ECmd[4] = Mathf.Lerp(ECmd[4], output, 1 - SmoothingSlider.val);
            ETarget[4] = ECmd[4];
        }

        private float SM(float m, float t)
        {
            if (m > t)
                return 0f;

            float delta = t - m;
            return (delta / t);
        }

        private float SX(float x)
        {
            return 1f - (float)Math.Sqrt((double)x);
        }

        private ValueTimed _l0VT = new ValueTimed();
        private ValueTimed _r1VT = new ValueTimed();
        private ValueTimed _r2VT = new ValueTimed();
        private ValueSplit _r12S = new ValueSplit();

        public float TestUpdateMotion(IMotionSource motionSource)
        {
            return RunMotion(motionSource, true);
        }

        public float UpdateMotion(IMotionSource motionSource)
        {
            return RunMotion(motionSource, false);
        }

        public float RunMotion(IMotionSource motionSource, bool testMode)
        {
            var length = motionSource.ReferenceLength * motionSource.ReferenceLengthMultiplier * ReferenceLengthScaleSlider.val;
            var radius = motionSource.ReferenceRadius * motionSource.ReferenceRadiusMultiplier * ReferenceRadiusScaleSlider.val;
            var referenceEnding = motionSource.ReferencePosition + motionSource.ReferenceUp * length;
            var diffPosition = motionSource.TargetPosition - motionSource.ReferencePosition;

            if (testMode)
            {

                //float dpm = diffPosition.magnitude;
                float maxRange = length * 1.2f;
                //if (dpm > maxRange)
                //    return -1f; // way too far to register

                // motionSource.ReferencePosition - base of penis
                // referenceEnding - end of penis

                // motionSource.TargetPosition - start of the female motion source

                // check nearest point within length of the target
                //Vector3 pntEnd = NearestPointOnLine(motionSource.TargetPosition, motionSource.TargetUp, referenceEnding);
                //Vector3 diff = pntEnd - motionSource.TargetPosition;
                Vector3 pntEnd = NearestPointOnLine(motionSource.ReferencePosition, motionSource.ReferenceUp, motionSource.TargetPosition);
                Vector3 diff = pntEnd - motionSource.ReferencePosition;
                // DebugDraw.DrawLine(pntEnd, motionSource.TargetPosition, Color.magenta);
                float dm = diff.magnitude;
                if (dm > maxRange)
                    return -1f;

                // check within radius (as done during tracking - copy-pasted code from below)
                float tR = Mathf.Clamp(Vector3.Dot(motionSource.TargetPosition - motionSource.ReferencePosition, motionSource.ReferenceUp), 0f, maxRange);
                Vector3 closestPointR = motionSource.ReferencePosition + motionSource.ReferenceUp * tR;
                // DebugDraw.DrawLine(closestPointR, motionSource.TargetPosition, Color.magenta);
                if (Vector3.Magnitude(closestPointR - motionSource.TargetPosition) > radius)
                    return -1f;

                diff.Normalize();
                float d = Vector3.Dot(diff, motionSource.TargetUp);
                if (d > 0f)
                {
                    // DebugDraw.DrawLine(pntEnd, motionSource.TargetPosition, Color.blue);
                    // SuperController.LogMessage((1f - Math.Abs(Mathf.Cos((dm / length) * Mathf.PI))).ToString());
                    // SuperController.LogMessage($"{dm} / {length} / {maxRange}");
                    if (dm > length)
                    {
                        // dm is within 1.0 to 1.2 of length here. That's not quite within tracking distance, but getting close!
                        float p = (maxRange - dm) / (maxRange - length);
                        // SuperController.LogMessage((p * 0.8f + 0.2f).ToString());
                        return p * 0.8f + 0.2f; // returning 0.2f to 1.0f here (where 0.2 will be when dmp is at max range)
                    }
                    return 1f;
                }
                return -1f;
            }

            for (var i = 0; i < 5; i++)
                DebugDraw.DrawCircle(Vector3.Lerp(motionSource.ReferencePosition, referenceEnding, i / 4.0f), motionSource.ReferenceUp, motionSource.ReferenceRight, Color.grey, radius);

            float t = Mathf.Clamp(Vector3.Dot(motionSource.TargetPosition - motionSource.ReferencePosition, motionSource.ReferenceUp), 0f, length);
            Vector3 closestPoint = motionSource.ReferencePosition + motionSource.ReferenceUp * t;

            if (Vector3.Magnitude(closestPoint - motionSource.TargetPosition) <= radius)
            {
                float xTarget0 = 1f;
                float xTarget1 = 0f;
                float xTarget2 = 0f;

                if (diffPosition.magnitude > 0.0001f)
                {
                    var diffEnding = motionSource.TargetPosition - referenceEnding;
                    bool aboveTarget = (Vector3.Dot(diffPosition, motionSource.TargetUp) < 0 && Vector3.Dot(diffEnding, motionSource.TargetUp) < 0)
                                        || Vector3.Dot(diffPosition, motionSource.ReferenceUp) < 0;

                    xTarget0 = 1f - Mathf.Clamp01((closestPoint - motionSource.ReferencePosition).magnitude / length);
                    if (aboveTarget)
                        xTarget0 = xTarget0 > 0.2f ? 1 : 0;

                    // SuperController.LogMessage(aboveTarget.ToString() + " / " + xTarget0.ToString());

                    var diffOnPlane = Vector3.ProjectOnPlane(diffPosition, motionSource.ReferencePlaneNormal);
                    var yOffset = Vector3.Project(diffOnPlane, motionSource.ReferenceRight);
                    var zOffset = Vector3.Project(diffOnPlane, motionSource.ReferenceForward);
                    xTarget1 = yOffset.magnitude * Mathf.Sign(Vector3.Dot(yOffset, motionSource.ReferenceRight));
                    xTarget2 = zOffset.magnitude * Mathf.Sign(Vector3.Dot(zOffset, motionSource.ReferenceForward));
                }

                var correctedRight = Vector3.ProjectOnPlane(motionSource.TargetRight, motionSource.ReferenceUp);
                if (Vector3.Dot(correctedRight, motionSource.ReferenceRight) < 0)
                    correctedRight -= 2 * Vector3.Project(correctedRight, motionSource.ReferenceRight);

                float rTarget0 = Vector3.SignedAngle(motionSource.ReferenceRight, correctedRight, motionSource.ReferenceUp) / 180;
                float rTarget1 = -Vector3.SignedAngle(motionSource.ReferenceUp, Vector3.ProjectOnPlane(motionSource.TargetUp, motionSource.ReferenceForward), motionSource.ReferenceForward) / 90;
                float rTarget2 = Vector3.SignedAngle(motionSource.ReferenceUp, Vector3.ProjectOnPlane(motionSource.TargetUp, motionSource.ReferenceRight), motionSource.ReferenceRight) / 90;

                float td = Time.time - TargetSwappedTime;
                if (td < 0.5f)
                {
                    td = ((td + 0.03f) / 0.53f);

                    float t0 = xTarget0 - XTarget[0];
                    float t1 = xTarget1 - XTarget[1];
                    float t2 = xTarget2 - XTarget[2];

                    xTarget0 = XTarget[0] + t0 * td;
                    xTarget1 = XTarget[1] + t1 * td;
                    xTarget2 = XTarget[2] + t2 * td;

                    t0 = rTarget0 - RTarget[0];
                    t1 = rTarget1 - RTarget[1];
                    t2 = rTarget2 - RTarget[2];

                    rTarget0 = RTarget[0] + t0 * td;
                    rTarget1 = RTarget[1] + t1 * td;
                    rTarget2 = RTarget[2] + t2 * td;
                }

                XTarget[0] = xTarget0;
                XTarget[1] = xTarget1;
                XTarget[2] = xTarget2;

                RTarget[0] = rTarget0;
                RTarget[1] = rTarget1;
                RTarget[2] = rTarget2;

                ETarget[0] = OutputV0CurveEditorSettings.Evaluate(XTarget, RTarget);
                ETarget[1] = OutputA0CurveEditorSettings.Evaluate(XTarget, RTarget);
                ETarget[2] = OutputA1CurveEditorSettings.Evaluate(XTarget, RTarget);
                ETarget[3] = OutputA2CurveEditorSettings.Evaluate(XTarget, RTarget);

                if (_lastNoCollisionTime != null)
                {
                    _lastNoCollisionSmoothingEnabled = true;
                    _lastNoCollisionSmoothingStartTime = Time.time;
                    _lastNoCollisionSmoothingDuration = Mathf.Clamp(Time.time - _lastNoCollisionTime.Value, 0.5f, 2);
                    _lastNoCollisionTime = null;
                }

                return 1f;
            }
            else
            {
                if (_lastNoCollisionTime == null)
                    _lastNoCollisionTime = Time.time;

                return -1f;
            }
        }

        public static Vector3 NearestPointOnLine(Vector3 linePnt, Vector3 lineDir, Vector3 pnt)
        {
            lineDir.Normalize();//this needs to be a unit vector
            var v = pnt - linePnt;
            var d = Vector3.Dot(v, lineDir);
            return linePnt + lineDir * d;
        }

        public void Dispose() => Dispose(true);
        protected virtual void Dispose(bool disposing) { }

        public virtual void OnSceneChanging() => _isLoading = true;
        public virtual void OnSceneChanged()
        {
            _lastNoCollisionSmoothingEnabled = true;
            _lastNoCollisionSmoothingStartTime = Time.time;
            _lastNoCollisionSmoothingDuration = 2;
            _lastNoCollisionTime = null;
            TargetSwappedTime = Time.time;

            _isLoading = false;
        }

        private class ValueTimed
        {
            public float High = -9999f;
            public float Low = 9999f;
            public float Value = 0f;
            public float LastTime = 0f;
            public float Magnitude = 0f;

            public float TimeHigh = 0f;
            public float TimeLow = 0f;

            public void Update(float v)
            {
                float d = Time.time - LastTime;
                LastTime = Time.time;
                Value = v;
                if (High < Value)
                {
                    High = Value;
                    TimeLow = Time.time;
                    TimeHigh = Time.time;
                }
                if (Low > Value)
                {
                    Low = Value;
                    TimeLow = Time.time;
                    TimeHigh = Time.time;
                }

                float s = d / 0.5f;
                if (Time.time - TimeHigh > 0.3f)
                {
                    High = Math.Max(Value, High - s);
                }
                if (Time.time - TimeLow > 0.3f)
                {
                    Low = Math.Min(Value, Low + s);
                }

                Magnitude = High - Low;
            }

        }

        public class ValueSplit
        {
            /// <summary>
            /// Percentage A
            /// </summary>
            public float PerA = 0f;

            /// <summary>
            /// Percentage B
            /// </summary>
            public float PerB = 0f;

            public float LastTime = 0f;

            /// <param name="mA">Magnitude A</param>
            /// <param name="mB">Magnitude B</param>
            public void Update(float mA, float mB)
            {
                float d = Time.time - LastTime;
                LastTime = Time.time;

                if (Math.Abs(mA - mB) < 0.001f)
                    return;

                float s = d / 2f;
                if (mA > mB)
                {
                    PerA = Math.Min(1f, PerA + s);
                } 
                else
                {
                    PerA = Math.Max(0f, PerA - s);
                }
                PerB = 1f - PerA;
            }
        }

    }
}
