using System;
using System.Linq;
using ToySerialController.Config;
using ToySerialController.MotionSource;
using DebugUtils;
using ToySerialController.Utils;
using ToySerialController.UI;

namespace ToySerialController
{
    public partial class Plugin : MVRScript
    {
        public static MVRScript Singleton { get; private set; }

        public static readonly string PluginName = "ToySerialController_VAMLaunch";
        public static readonly string PluginAuthor = "Blazedust (and others)"; // Thanks for Yoooi, Zengineer, hazmhox for the original plugins.
        public static readonly string PluginDir = $@"Custom\Scripts\{PluginName.Replace(" ", "")}";

        public static bool DebugEnabled = false;

        private IDevice _device;
        private IMotionSource _motionSource;
        private bool _initialized;
        private bool _isLoading;
        private int _physicsIteration;
        private int _loadDefaultConfig = 10;

        private SuperController Controller => SuperController.singleton;

        public override void Init()
        {
            Singleton = this;
            base.Init();

            UIManager.Initialize(this);
        }

        public override void InitUI()
        {
            base.InitUI();
            if (UITransform == null)
                return;

            try
            {
                CreateUI();
                _initialized = true;
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        protected void Update()
        {
            if (SuperController.singleton.isLoading)
                ComponentCache.Clear();

            if (!_initialized || SuperController.singleton.isLoading)
                return;

            if (_loadDefaultConfig > 0)
            {
                _loadDefaultConfig--;
                if (_loadDefaultConfig <= 0)
                {
                    try
                    {
                        LoadDefaultConfig();
                    }
                    catch
                    {
                        // ignore
                    }
                }
            }

            DebugDraw.Draw();
            DebugDraw.Enabled = DebugDrawEnableToggle.val;
            DebugEnabled = DebugDrawEnableToggle.val;
            _physicsIteration = 0;
        }

        protected void FixedUpdate()
        {
            if (!_initialized)
            {
                return;
            }

            var isLoading = SuperController.singleton.isLoading;
            if (!_isLoading && isLoading)
                OnSceneChanging();
            else if (_isLoading && !isLoading)
                OnSceneChanged();
            _isLoading = isLoading;

            if (_physicsIteration == 0)
                DebugDraw.Clear();

            UpdateDevice();

            if (_physicsIteration == 0)
                DebugDraw.Enabled = false;

            _physicsIteration++;
        }

        private void LoadDefaultConfig()
        {
            try
            {
                try
                {
                    var defaultPath = Controller.GetFilesAtPath(PluginDir, "*.json").FirstOrDefault(s => s.EndsWith("default.json"));
                    if (defaultPath != null)
                        ConfigManager.LoadConfig(defaultPath, this, false);
                }
                catch (Exception e)
                {
                    SuperController.LogError("Exception caught: " + e);
                }
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        private void UpdateDevice()
        {
            try
            {
                if (_device != null)
                {
                    var motionValid = _motionSource?.Update(_device.TestUpdateMotion) == true;
                    _device.Update(motionValid ? _motionSource : null, _outputTarget);
                    if (ToySerialController.Plugin.DebugEnabled) 
                        DeviceReportText.val = _device.GetDeviceReport() ?? string.Empty;
                }
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        protected void OnSceneChanging()
        {
            _device?.OnSceneChanging();
            _motionSource?.OnSceneChanging();
            _outputTarget?.OnSceneChanging();
        }

        protected void OnSceneChanged()
        {
            _device?.OnSceneChanged();
            _motionSource?.OnSceneChanged();
            _outputTarget?.OnSceneChanged();
        }

        protected void OnDestroy()
        {
            try
            {
                DebugDraw.Clear();
                _device?.Dispose();
                _outputTarget?.Dispose();
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }
    }
}